#!/usr/bin/env python2
"""
All functionality related to filtering out Git tags against
Semantic Versioning or custum rules
"""

import re
import logging
import semver


class TagFilter(object):
    """
    Class facilitating validating and filtering versions against
    SemVer rules or custom, user-provided version scheme
    """
    def __init__(self, pattern=None, prefix=None):
        self.prefix = prefix
        if pattern is None:
            self.matches = self.is_semantic_version
        else:
            self.matches = self.is_custom_version
            self.custom_pattern = re.compile(pattern)

    def is_semantic_version(self, value, prefix=None):
        """
        Verify if given value is a Semantic Version (as defined by semver.org)

        :param value: input string
        :param prefix: optional prefix (e.g. 'v') to be stripped off
                       the input string. If omitted, use prefix given
                       at object creation time.
        :return: True or False
        """
        logging.debug("Validate string '%s' against SemVer rules", value)
        version_string = self.strip_prefix(value, prefix)
        try:
            semver.parse(version_string)
        except ValueError:
            return False
        return True

    def is_custom_version(self, value, prefix=None):
        """
        Verify if given value matches a custom regular expression
        specified by a command-line option.

        :param value: input string
        :param prefix: optional prefix (e.g. 'v') to be stripped off
                       the input string
        :return: True or False
        """
        logging.debug("Validate string '%s' against custom pattern '%s'",
                      value, self.custom_pattern.pattern)
        version_string = self.strip_prefix(value, prefix)
        retval = False
        if self.custom_pattern.match(version_string):
            retval = True
        return retval

    def matching_only(self, versions, prefix=None):
        """
        Return only the list of versions (tags) that match the configured
        versioning rules

        :param versions: list of strings to check
        :param prefix: prefix of the version string; if present, will be
                       ignored before checking if string is a semantic version
        :return: list containing only those strings which match SemVer.org
                 specification; may be empty.
        """
        return [x for x in versions if
                self.matches(x, prefix=prefix)]

    def strip_prefix(self, value, prefix=None):
        """

        :param prefix: optional prefix (e.g. 'v') to be stripped off
                       the input string. If omitted, use prefix given
                       at object creation time.
        :param value: version string
        :return:
        """
        if prefix is None:
            prefix = self.prefix

        version_string = value
        if prefix and value.startswith(prefix):
            version_string = value[len(prefix):]
        return version_string



#!/usr/bin/env python2
"""
Unit tests for version validation and matching
"""

from gcg.tag_filter import TagFilter


def test_semantic_version_ok():
    """Various correct semantic version strings"""
    obj = TagFilter()
    assert obj.matches('1.2.0')
    assert obj.matches('1.2.0-3')
    assert obj.matches('foo1.2.0', prefix='foo')

    assert TagFilter(prefix='v').matches('v1.2.0')
    assert TagFilter(prefix='foo').matches('foo1.2.0')


def test_semantic_version_failures():
    """Various invalid semantic version strings, or with invalid
    prefix"""
    obj = TagFilter()
    assert not obj.matches('foo1.2.0', prefix='oof')
    assert not obj.matches('1')
    assert not obj.matches('1.2')
    assert not obj.matches('v1.2.3')


def test_custom_tag_pattern_success():
    """
    Simple success scenarios
    """
    assert TagFilter(r'v(\d+)\.(\d+)').matches('v123.123')
    assert TagFilter('v(\\d+)\\.(\\d+)').matches('v123.123')


def test_custom_tag_pattern_fails():
    """Simple failing scenarios"""
    assert not TagFilter(r'^(\d+)\.(\d+)$').matches('v123.123')
    assert not TagFilter(r'^(\d+)\.(\d+)$').matches('1.2.3')


def test_prefix_is_optional():
    """Lack of prefix is still considered a valid version"""
    assert TagFilter(prefix='foo').matches('foo1.2.0')
    assert TagFilter(r'^(\d+)\.(\d+)$', 'v').matches('v123.123')


def test_filter_semvers_only():
    """Test only semver-compliant versions are returned by only_semvers"""
    input_list = [
        'v1.2',
        '1.4',
        '0.1.2',
        '99.88.77-654321',
        '22.33.1-YADA',
        '1.2.3_5',
    ]
    expected = [
        '0.1.2',
        '99.88.77-654321',
        '22.33.1-YADA',
    ]
    assert expected == TagFilter().matching_only(input_list)


def test_matching_only_with_prefix():
    """Test only semver-compliant versions are returned by only_semvers
    while using a prefix. The prefix is optional."""
    input_list = [
        'v1.2',
        '1.4',
        '0.1.2',
        'v0.1.2',
        'v99.88.77-654321',
        '22.33.1-YADA',
        '1.2.3_5',
    ]
    expected = [
        '0.1.2',
        'v0.1.2',
        'v99.88.77-654321',
        '22.33.1-YADA',
    ]
    assert expected == TagFilter(prefix='v').matching_only(input_list)
    assert expected == TagFilter().matching_only(input_list, prefix='v')


def test_matching_only_custom():
    """
    Filter out unwanted items from the list, with and without
    the prefix
    """

    assert TagFilter(r'^(\d+)\.(\d+)$').matching_only([
        'v123.123',
        '01.02',
        '1.2.3',
        '1.2.3-4+bld5']) == ['01.02']

    assert TagFilter(r'^(\d+)\.(\d+)$', prefix='v').matching_only([
        'v123.123',
        '01.02',
        '1.2.3',
        '1.2.3-4+bld5']) == ['v123.123', '01.02']



#!/usr/bin/env python2.7
# encoding: utf-8

'''
Test cases for simple functions which don't fit anywhere else (as of yet).
'''
import argparse
import collections
import logging
import shutil

import os
import tempfile

import gcg


class TestHelperFunctions(object):
    """Tests of remaining helper functions"""

    # pylint: disable=unused-argument,missing-docstring
    # pylint: disable=attribute-defined-outside-init
    def setup_method(self, test_method):
        self.temp_dir = tempfile.mkdtemp()

    # pylint: disable=unused-argument,missing-docstring
    def teardown_method(self, test_method):
        shutil.rmtree(self.temp_dir)

    @staticmethod
    def test_loglevels_mapping():
        from gcg.entrypoint import log_level_from_verbosity

        assert logging.WARNING == log_level_from_verbosity(0)
        assert logging.WARNING == log_level_from_verbosity(-1)
        assert logging.WARNING == log_level_from_verbosity(3)

        assert logging.INFO == log_level_from_verbosity(1)
        assert logging.DEBUG == log_level_from_verbosity(2)

    def test_print_changelog_file_ok(self):
        """print_changelog() succesfully creates output file"""
        ofile = os.path.join(self.temp_dir, 'output.txt')
        entries = collections.OrderedDict()
        headers = []
        gcg.entrypoint.print_changelog(
            entries, headers, output_format='rpm', output_file=ofile)
        assert os.path.exists(ofile)

    # FIXME: fix this UT TO ACTUALLY VALIDATE THE OUTPUT, i think
    def test_print_changelog_use_stdout(self):
        # repo = git.Repo.init(self.tmp_dir)
        logging.debug("Go away, pylint! %s", self)
        entries = collections.OrderedDict()
        fake_commit = argparse.Namespace()
        fake_commit.date_rpm = 'baz'
        fake_commit.message = 'msg'
        fake_commit.hexsha = 'a1a2a3a4a5a6'
        entries['foo'] = [fake_commit]
        headers = {
            'foo': 'bar'
        }
        gcg.entrypoint.print_changelog(entries, headers, output_format='rpm')

        # print("OUTPUT: {}".format(mock_stdout.getvalue()))



#!/usr/bin/env python2

"""
Unit and component tests for the main module
"""

from __future__ import print_function
import os
import shutil
import tempfile

import gcg
import gcg.entrypoint
import gcg.errors as err
from tests.helpers.gitrepo import prepare_git_repo, AUTHOR


class TestMain(object):
    """Unit test suite for main functionality. May be split into
    smaller parts with time."""

    # pylint: disable=unused-argument,missing-docstring
    # pylint: disable=attribute-defined-outside-init,duplicate-code
    def setup_method(self, test_method):
        self.tmp_dir = tempfile.mkdtemp()

    # pylint: disable=unused-argument,missing-docstring
    def teardown_method(self, test_method):
        shutil.rmtree(self.tmp_dir)

    @staticmethod
    def test_main_rpm():
        """Generate changelog for THIS repo. Illustrative purposes only"""
        test_path = os.path.dirname(os.path.abspath(__file__))
        repo_path = os.path.join(test_path, '..')
        assert err.SUCCESS == gcg.entrypoint.main([
            'xyz', '-p', repo_path, '-O', 'rpm'])

    @staticmethod
    def test_main_deb():
        """Generate changelog for THIS repo. Illustrative purposes only"""
        test_path = os.path.dirname(os.path.abspath(__file__))
        repo_path = os.path.join(test_path, '..')
        assert err.SUCCESS == gcg.entrypoint.main([
            'xyz', '-p', repo_path, '-O', 'deb', '-n', 'FOOBAR',
            '--deb-distribution', 'xenial'])

    def test_main_rpm_fakerepo(self):
        """
        Generate a changelog from a well known and defined repository
        which we control ourselves.
        """

        input_messages = [
            "Initial commit",
            "And then a second commit",
            "ISSUE-1 And a third commit\nThis time with a Jira ticket",
        ]

        (path, repo, client) = prepare_git_repo(
            self.tmp_dir, messages=input_messages)
        assert not repo.bare
        assert not client.tag()

        out_file = os.path.join(self.tmp_dir, 'outfile')
        assert err.SUCCESS == gcg.entrypoint.main([
            'xyz', '-p', path, '-O', 'rpm', '-o', out_file])
        assert os.path.exists(out_file)

        lines = [line.rstrip('\n') for line in open(out_file)]
        assert lines[0].endswith(
            '{} <{}> - current'.format(AUTHOR.name, AUTHOR.email))
        assert lines[1].startswith('- ISSUE-1 And a third commit')
        assert lines[2].startswith('- And then a second commit')
        assert lines[3].startswith('- Initial commit')

    def test_main_deb_fakerepo(self):
        """
        Generate a changelog from a well known and defined repository
        which we control ourselves - DEB
        """
        input_messages = [
            "First",
            "Second",
        ]

        (path, repo, client) = prepare_git_repo(
            self.tmp_dir, messages=input_messages)
        assert not repo.bare
        assert client

        out_file = os.path.join(self.tmp_dir, 'outfile')
        assert err.SUCCESS == gcg.entrypoint.main([
            'xyz', '-p', path, '-O', 'deb', '-o', out_file, '-n', 'foopkg',
            '-D', 'xenial'])
        assert os.path.exists(out_file)

        lines = [line.rstrip('\n') for line in open(out_file)]
        lines = filter(len, lines)
        assert lines[0] == 'foopkg (current) xenial; urgency=low'
        assert lines[1].startswith('  * Second')
        assert lines[2].startswith('  * First')

    def test_main_deb_custom_urgency(self):
        """
        Verify the deb urgency is recorded as appropriate
        """
        input_messages = [
            "First",
        ]

        (path, repo, client) = prepare_git_repo(
            self.tmp_dir, messages=input_messages)
        assert not repo.bare
        assert client

        out_file = os.path.join(self.tmp_dir, 'outfile')
        assert err.SUCCESS == gcg.entrypoint.main([
            'xyz', '-p', path, '-O', 'deb', '-o', out_file, '-n', 'package',
            '-D', 'xenial', '-U', 'high'])
        assert os.path.exists(out_file)

        lines = [line.rstrip('\n') for line in open(out_file)]
        lines = filter(len, lines)
        assert lines[0] == 'package (current) xenial; urgency=high'
        assert lines[1].startswith('  * First')

    def test_main_only_jira_issues(self):
        """
        Validate that with -b option, only commits containing
        JIRA issues (default) are included
        """

        input_messages = [
            "Initial commit",
            "And then a second commit",
            "ISSUE-1 And a third commit\nThis time with a Jira ticket",
        ]

        # pylint: disable=unused-variable
        (path, repo, client) = prepare_git_repo(
            self.tmp_dir, messages=input_messages)

        out_file = os.path.join(self.tmp_dir, 'outfile')
        assert err.SUCCESS == gcg.entrypoint.main([
            'xyz', '-p', path, '-O', 'rpm', '-o', out_file, '-b'])
        assert os.path.exists(out_file)

        lines = [line.rstrip('\n') for line in open(out_file)]
        assert len(lines) == 3
        assert lines[0].endswith('> - current')
        assert lines[1].startswith('- ISSUE-1 And a third commit')

    def test_only_gitlab_issues(self):
        """
        Validate that with -b option, only commits containing
        gitlab issues by passing in a -B and a custom pattern
        """

        input_messages = [
            "Initial commit",
            "Issue #312 Resolve the bug",
        ]

        # pylint: disable=unused-variable
        (path, repo, client) = prepare_git_repo(
            self.tmp_dir, messages=input_messages)

        out_file = os.path.join(self.tmp_dir, 'outfile')
        assert err.SUCCESS == gcg.entrypoint.main([
            'xyz', '-p', path, '-O', 'rpm', '-o', out_file, '-b',
            '-B', r'Issue #\d+\b'])

        lines = [line.rstrip('\n') for line in open(out_file)]
        assert len(lines) == 3
        assert lines[0].endswith('> - current')
        assert lines[1].startswith('- Issue #312 Resolve the bug')

    def test_merge_repo(self):
        input_messages = ["1", "2", "3"]

        (path, repo, client) = prepare_git_repo(
            self.tmp_dir, messages=input_messages)
        test_branch = repo.create_head('test_branch', 'HEAD~1')

        repo.head.reference = test_branch
        repo.index.commit('4')
        repo.head.reference = repo.heads.master
        repo.index.commit('5')
        client.merge('--no-ff', '-m 6', 'test_branch')

        out_file = os.path.join(self.tmp_dir, 'outfile')
        assert err.SUCCESS == gcg.entrypoint.main([
            'xyz', '-p', path, '-O', 'rpm', '-o', out_file])
        assert os.path.exists(out_file)

        lines = [line.rstrip('\n') for line in open(out_file)]
        lines = filter(len, lines)
        print(lines)

        count = len(lines)
        for i in range(1, count):
            assert lines[count-i].startswith('- {}'.format(i))



#!/usr/bin/env python2

"""Unit tests for various combinations of main() parameters"""

import logging
import shutil

import tempfile
import unittest

import git
from mock import patch

import gcg.entrypoint
import gcg.errors as err


class TestUserInputs(unittest.TestCase):
    """Unit test suite for invalid user inputs"""
    def setUp(self):
        logging.basicConfig(level=logging.FATAL)
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    @staticmethod
    def test_main_missing_debpkgname():
        """Missing package name for DEB output type"""
        assert err.INVALID_INPUT == gcg.entrypoint.main([
            'xyz', '-O', 'deb'])

    @staticmethod
    def test_main_missing_debdistro():
        """Missing debian distribution for DEB output type"""
        assert err.INVALID_INPUT == gcg.entrypoint.main([
            'xyz', '-O', 'deb', '-n', 'pkgname'])

    @staticmethod
    def test_main_repo_invalid_path():
        """Path does not exist"""
        assert err.REPO_PATH_INVALID == gcg.entrypoint.main([
            'xyz', '-p', '/doesNot3xist', '-O', 'rpm'])

    def test_main_path_is_not_repo(self):
        """A directory given is not a Git repository"""
        assert err.REPO_PATH_NOT_REPO == gcg.entrypoint.main([
            'xyz', '-p', self.tmp_dir, '-O', 'rpm'])

    @staticmethod
    @patch('git.Repo')
    def test_main_ref_inv_upper_limit(mock_git_repo):
        """Invalid upper limit specified by reference"""
        mock_git_repo().commit.side_effect = git.BadName('Bang')
        assert err.INVALID_VCS_LIMITS == gcg.entrypoint.main([
            'xyz', '--until', 'NonExistantTag', '-O', 'rpm'])

    @staticmethod
    @patch('git.Repo')
    def test_main_sha_inv_upper_limit(mock_git_repo):
        """Invalid upper limit specified by commit hash"""
        mock_git_repo().commit.side_effect = ValueError('Boom')
        assert err.INVALID_VCS_LIMITS == gcg.entrypoint.main([
            'xyz', '--until', 'a1a2a2a3a4a5a6a', '-O', 'rpm'])

    @staticmethod
    @patch('git.Repo')
    def test_main_ref_inv_low_limit(mock_git_repo):
        """Invalid lower limit specified by reference"""
        mock_git_repo().commit.side_effect = git.BadName('Bang')
        assert err.INVALID_VCS_LIMITS == gcg.entrypoint.main([
            'xyz', '--since', 'NonExistantTag', '-O', 'rpm'])

    @staticmethod
    @patch('git.Repo')
    def test_main_sha_inv_low_limit(mock_git_repo):
        """Invalid lower limit specified by commit hash"""
        mock_git_repo().commit.side_effect = ValueError('Boom')
        assert err.INVALID_VCS_LIMITS == gcg.entrypoint.main([
            'xyz', '--since', 'a1a2a2a3a4a5a6a', '-O', 'rpm'])



#!/usr/bin/env python2

"""Unit test suite for functions related to version handling"""

from argparse import Namespace
import gcg.jinja_filters


def test_commit_headline_ok():
    """
    basic success scenario of commit_headline filter
    """
    fake_commit = Namespace(message="""This is a   \

    multi-line message with spaces at the   end   \

    of some lines      \

    """)
    assert gcg.jinja_filters.commit_headline(fake_commit) == "This is a"



#!/usr/bin/env python2

"""
gcg stands for Git Changelog Generator
This module can be used to traverse the Git version tree
between specified commits (by default, from HEAD to the first reachable
commit) and produce a DEB or RPM changelog output, treating
tags as release points.

Note that only tags conforming to semver.org policies are treated
as releases.

"""
from __future__ import print_function

import argparse
import collections
import email.utils
import logging
import os
import re
import sys
import time

import git
import jinja2
import jinja2.environment

import gcg.errors as err
from gcg.tag_filter import TagFilter
from gcg.jinja_filters import commit_headline


def parse_args(argv):
    """Parse user's command-line parameters

    :param argv: User arguments
    :returns: object with configuration (as provided by argparse)
    """

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Generate changelogs out of Git history.",
        epilog="""
        FIXME: timezone manipulations
        FIXME: potential time conflicts (merging in old changes
        to a new release).
        TODO: better test coverage
        """
    )
    parser.add_argument(
        '-p', '--path',
        help="""Git repository path""",
        type=str, action="store",
        default=os.getcwd(),
    )
    parser.add_argument(
        '-O', '--output-format',
        help="""Output changelog format""",
        type=str, action="store", choices=['rpm', 'deb'],
        required=True
    )
    parser.add_argument(
        '-o', '--output-file',
        help="""Output file path. Uses standard output when omitted""",
        type=str, action="store",
    )
    parser.add_argument(
        '-u', '--until',
        help="Most recent (top) commit to include in the "
             "changelog. This is the commit from which history discovery"
             "begins",
        type=str, action="store",
        default='HEAD'
    )
    parser.add_argument(
        '-s', '--since',
        help="Oldest commit to include in the changelog",
        type=str, action="store"
    )
    parser.add_argument(
        '-c', '--current-version',
        help="""Version to be assigned to the revisions between
        the last tagged version and specified <end-commit> (e.g. HEAD).
        For example: .... (FIXME)""",
        type=str, action="store",
        default='current'
    )
    parser.add_argument(
        '-x', '--exclude-merges',
        help="Filter out the merge commits out of changelog.",
        action="store_true",
    )
    parser.add_argument(
        '-b', '--bug-tracking-only',
        help="Process only commits which have a bug tracking"
             "issue reference in the message (e.g. a Jira ticket).",
        action="store_true",
    )
    parser.add_argument(
        '-B', '--bug-tracking-pattern',
        help="Provide the pattern (Python regexp) to check against commit"
             "messages when using --bug-tracking-only option."
             "The default value matches a standard Jira issue reference.",
        action="store", default=r'\b(([A-Z]{1,10})-?)[A-Z]+-[1-9]+\d*\b'
    )
    parser.add_argument(
        '-t', '--prefer-tags',
        help="""Retrieve the release information (date and release author)
        from tags. When false (the default), the information will
        be extracted from the most recent commit preceding the tag
        or commit specified by --until option.""",
        action="store_true",
    )
    parser.add_argument(
        '-T', '--custom-tag-pattern',
        help="""Provide the pattern (Python regexp) to match against Git tags.
        Tags which don't match the pattern will not be considered as releases.
        When omitted, tag names will be validated against Semantic
        Versioning (semver.org) rules.""",
        action="store"
    )
    parser.add_argument(
        '-D', '--deb-distribution',
        help="""Distribution for the deb package.
        NOTE: mandatory when the output format (-O) is 'deb'.""",
        action="store"
    )
    parser.add_argument(
        '-n', '--deb-package-name',
        help="""Package name (in case it's a Debian package).
        NOTE: mandatory when the output format (-O) is 'deb'.""",
        action="store",
    )
    parser.add_argument(
        '-U', '--deb-urgency',
        help="""Default urgency value for Debian releases.
        NOTE: this is a single value which will be populated across
        the whole changelog, for all releases.""",
        action="store", default='low',
        choices=['low', 'medium', 'high', 'emergency', 'critical']
    )
    parser.add_argument(
        '-v', '--verbose',
        help="Verbosity level. Specify twice for debug logs as well.",
        action="count", default=0
    )

    options = parser.parse_args(argv)
    if options.output_format == 'deb':
        if options.deb_package_name is None:
            raise ValueError("For 'deb', the package name (-n) is mandatory")
        if options.deb_distribution is None:
            raise ValueError("For 'deb', the --deb-distribution is mandatory")

    return options


def get_commit_tags(client, commit):
    """
    Retrieve tags pointing at a specific commit/revision
    :param client: an initialized git.Git() object to query with
    :param commit:
    :return: list of strings containing tags pointing at specified
             revision; may be empty if there's no tag at this revision
    """
    retval = []
    (status, output, errors) = client.tag(
        '--points-at', commit.hexsha, with_extended_output=True)
    if status:
        logging.error(errors)
    else:
        if output:
            retval = output.split('\n')
    return retval


def collate_entry_header_data(repo, entries, options):
    """
    :param repo:
    :param entries: an OrderedDict object with keys being tag names (string)
                    and values being a list of git.Commit objects
    :param options: user input parameters
    :return: dictionary; k: tag name, v: object with properties
                datetime
    """
    retval = {}
    for version in entries:
        if not version or (entries[version] and not options.prefer_tags):
            hdr = log_entry_header_from_commit(entries[version][0])
        else:
            logging.info("Retrieving details of tag '%s'", version)
            hdr = log_entry_header_from_tag(repo.tags[version].tag)
        hdr['version'] = version or options.current_version
        hdr['deb_urgency'] = options.deb_urgency
        hdr['deb_distro'] = options.deb_distribution
        hdr['deb_name'] = options.deb_package_name
        retval[version] = hdr
    return retval


def log_entry_header_from_commit(the_commit):
    """
    Retrieve information for the release line from the commit
    :param the_commit: a git.Commit object
    :return: dictionary
    """
    # FIXME: how about timezones, dear people?
    hdr = dict()
    hdr['date'] = time.gmtime(the_commit.authored_date)
    hdr['date_rpm'] = time.strftime('%a %b %d %Y', hdr['date'])
    hdr['date_deb'] = email.utils.formatdate(the_commit.authored_date)
    hdr['author'] = str(the_commit.author.name)
    hdr['email'] = str(the_commit.author.email)
    return hdr


def log_entry_header_from_tag(tag):
    """
    Retrieve information for the release line from the tag
    :param tag: a git.Tag object
    :return: dictionary
    """
    # FIXME: how about timezones, dear people?
    hdr = dict()
    authored_date = tag.tagged_date
    hdr['date'] = time.gmtime(authored_date)
    hdr['date_rpm'] = time.strftime('%a %b %d %Y', hdr['date'])
    hdr['date_deb'] = email.utils.formatdate(authored_date)
    hdr['author'] = str(tag.tagger.name)
    hdr['email'] = str(tag.tagger.email)
    return hdr


def log_level_from_verbosity(count):
    """Convert verbosity (-v) count into an appropriate
    log level
    :param count: integer; number of -v flags used
    :returns: a log level value
    """
    retval = logging.WARNING
    if count == 1:
        retval = logging.INFO
    elif count == 2:
        retval = logging.DEBUG
    return retval


def init_jinja_env(external_template_dir=None):
    """
    Initialize Jinja environment:
    - load templates embedded in the gcg module
    - optionally load from an external template directory
    - register custom Jinja filters
    :param external_template_dir: optional; when provided, templates from given
                                  directory will be looked up first
    :return: the newly created Jinja environment
    """
    loaders = []
    if external_template_dir:
        loaders.append(jinja2.FileSystemLoader(external_template_dir))
    loaders.append(jinja2.PackageLoader('gcg'))

    env = jinja2.Environment(
        loader=jinja2.ChoiceLoader(loaders)
    )
    env.filters['commit_headline'] = commit_headline
    return env


# pylint: disable=too-many-return-statements
def main(argv=None):
    """
    Main entry point
    :param argv: optional script parameters; default is to use sys.argv
    :return:
    """
    if argv is None:
        argv = sys.argv

    try:
        options = parse_args(argv[1:])
    except ValueError as exc:
        print("Invalid input; details: {}".format(exc), file=sys.stderr)
        return err.INVALID_INPUT

    loglevel = log_level_from_verbosity(options.verbose)
    logging.basicConfig(
        level=loglevel, format="[%(levelname)s] %(message)s")

    try:
        repo = git.Repo(options.path)
    except ValueError as exc:
        logging.error("Invalid input; details: %s", exc)
        return err.INVALID_INPUT
    except git.NoSuchPathError as exc:
        logging.error("Path specified with '-p' does not exist; %s", exc)
        return err.REPO_PATH_INVALID
    except git.InvalidGitRepositoryError as exc:
        logging.error("Path specified with '-p' does not a Git repo make; %s",
                      exc)
        return err.REPO_PATH_NOT_REPO

    try:
        client = git.Git(options.path)
        upper_limit = resolve_commit_from_arguments(repo, options, 'until')
        lower_limit = resolve_commit_from_arguments(repo, options, 'since')
    except ValueError as exc:
        logging.error("Invalid input values for changelog scope; details: "
                      "%s.", exc)
        return err.INVALID_VCS_LIMITS

    try:
        entries = traverse_version_tree(
            client, repo, options, upper_limit, lower_limit)
        headers = collate_entry_header_data(repo, entries, options)

    except ValueError as exc:
        logging.error("Program aborted; details: %s.", exc)
        return err.PROCESSING_FAILED

    print_changelog(entries, headers, options.output_format,
                    options.output_file)

    return err.SUCCESS


def print_changelog(entries, headers, output_format, output_file=None):
    """
    Render the changelog report
    :param entries: typically commit objects
    :param headers: release headers (dictionary with release tag as key)
    :param output_format: string, e.g. 'rpm' or 'deb'
    :param output_file: optional; file name to write to. By default write
                        to standard output
    :return: n/a
    """
    environment = init_jinja_env()
    template = environment.get_template(output_format)
    output = template.render(entries=entries, headers=headers)
    if output_file:
        with open(output_file, 'w') as ofile:
            ofile.write(output)
    else:
        print(output)


def commit_filtered_out(commit, options, bugtracking_regex):
    """
    Decide if commit should be processed or not

    :param commit: Commit to be inspected
    :param options: command-line options
    :param bugtracking_regex: regular expression object to match
                              against the commit message. Only used if
                              options.bug_tracking_only evaluates to True
    :return: True / False
    """
    retval = False
    if options.exclude_merges and len(commit.parents) > 1:
        logging.info("Ignoring merge commit %s", commit)
        retval = True
    elif options.bug_tracking_only:
        if not bugtracking_regex.match(commit.message):
            logging.info("Ignoring commit %s (no bug reference)", commit)
            retval = True
    return retval


def repo_iterate(repo, high, low=None):
    """
    Return (yield) one commit by one, from 'high' to 'low'
    (or oldest reachable commit) and in reverse chronological order
    (see rev-list and git log commands)

    :param repo: an initialized git.Repo object
    :param high: top of the version tree to scan; typically tip of the branch,
                HEAD, but it can be any commit
    :param low: optional; the low end of the revision list. When unspecified,
                function will scan the history until the first reachable commit
    :return: a git.Commit object
    """

    refs = high if low is None else "{}..{}".format(high, low)

    for commit in repo.iter_commits(refs):
        yield commit


def traverse_version_tree(client, repo, options, upper_limit, lower_limit):
    """
    Scan the version tree and return the entries
    :param client: an initialized git.Git() object to query with
    :param repo: reference to the repository (git.Repo())
    :param options: command-line options as specified by the user
    :param upper_limit: the top commit from which scanning begins
    :param lower_limit: the bottom commit where scanning the version
                        tree ends; if None, scan will continue until
                        the very first reachable commit
    :return: an OrderedDict object with keys being tag names (string)
             and values being a list of git.Commit objects
    """
    entries = collections.OrderedDict()
    curr_entry = list()
    curr_tag = ''
    bugtracking_regexp = re.compile(options.bug_tracking_pattern, re.MULTILINE)
    tag_filter = TagFilter(None)

    for commit in repo_iterate(repo, upper_limit, lower_limit):
        logging.debug("Processing commit %s (%s)", commit.hexsha,
                      commit_headline(commit))
        tags = tag_filter.matching_only(get_commit_tags(client, commit))
        if tags:
            if curr_entry:
                entries[curr_tag] = curr_entry
            # TODO: handle more tags pointing to the same commit (somehow)
            curr_tag = tags[0]
            curr_entry = list()

        if not commit_filtered_out(commit, options, bugtracking_regexp):
            curr_entry.append(commit)
    if curr_entry:
        entries[curr_tag] = curr_entry
    return entries


def resolve_commit_from_arguments(repo, options, arg_name):
    """
    Resolve a reference to a commit (by branch name, label or sha)
    to an actual commit id
    :param arg_name: name of the command-line argument to resolve
    :param options: command-line options
    :param repo: reference to a repository (initialized)
    :returns: a commit id as string
    :raises: ValueError when the reference cannot be found within the repo
    """
    retval = None
    ref = getattr(options, arg_name)
    if ref is not None:
        invalid_commit_msg = "Value of argument --%s does not not resolve " \
                             "to a valid commit in this repository"
        try:
            retval = repo.commit(ref)
            logging.info("Commit referred to as '%s' (%s) resolved to %s",
                         arg_name, ref, retval.hexsha)
        except git.BadName as exc:
            logging.error(invalid_commit_msg, arg_name)
            raise ValueError(exc)
        except ValueError:
            logging.error(invalid_commit_msg, arg_name)
            raise
    return retval



#!/usr/bin/env python2

"""
Definition of exit codes for the main()
"""

SUCCESS = 0
ARGPARSE_FAILURE = 2
INVALID_INPUT = 10
REPO_PATH_INVALID = 11
REPO_PATH_NOT_REPO = 12
INVALID_VCS_LIMITS = 13
PROCESSING_FAILED = 20
UNHANDLED_EXCEPTION = 255
